using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Sockets;
using System.Threading;

namespace Smdn.Protocols.Imap4.Protocol {
  // Layers:
  //   Client.ImapClient
  //   Client.ImapClientAgent
  //   Session.ImapSessionAgent
  //   Session.ImapSession
  //   Transaction.ImapTransactionAgent
  //   Transaction.ImapTransaction/Transaction.ImapRequest/Transaction.ImapResponse
  // * Protocol.ImapConnection/Protocol.ImapProtocolCommand/Protocol.ImapProtocolResponse
  //   (IMAP4)

  public class ImapConnection {
    private class Receiver {
      public Receiver(Stream stream)
      {
        this.stream = stream;
      }

      public byte[] TryReceive()
      {
        lock (receivedData) {
          if (0 < receivedData.Count)
            return receivedData.Dequeue();
        }

        if (readEvent.WaitOne(0, false))
          BeginReceive();

        return null;
      }

      private void BeginReceive()
      {
        stream.BeginRead(buffer, 0, buffer.Length, EndReceive, this);
      }

      private void EndReceive(IAsyncResult ar)
      {
        // TODO: サーバからの切断の考慮
        var read = stream.EndRead(ar);

        if (read <= 0)
          return;

        var data = new byte[read];

        Array.Copy(buffer, data, read);

        lock (receivedData) {
          receivedData.Enqueue(data);
        }

        Trace.Dump(data);

        readEvent.Set();
      }

      private readonly Stream stream;
      private readonly AutoResetEvent readEvent = new AutoResetEvent(true);
      private readonly byte[] buffer = new byte[0x400];
      private readonly Queue<byte[]> receivedData = new Queue<byte[]>();
    }

    /*
    private class Sender {
      // TODO: impl
      public Sender(Stream stream)
      {
        this.stream = stream;
      }

      private readonly Stream stream;
    }
    */

    /*
     * properties
     */

    internal int Id {
      get { return id; }
    }

    public bool Connected {
      get { return connected; }
    }

    public bool IsIdling {
      get { return isIdling; }
      internal set { isIdling = value; }
    }

    public ImapIdlingResponseCallback IdlingCallback {
      get { return idlingCallback; }
      internal set { idlingCallback = value; }
    }

    public ImapServerInfo ServerInfo {
      get { return serverInfo; }
    }

    // for '5.4. Autologout Timer'
    public DateTime LastSentTime {
      get { return lastSentTime; }
    }

    public int QueuedResponseCount {
      get { return parsedResponse.Count; }
    }

    private ImapConnection(ImapUri serverUri, int id)
    {
      this.serverInfo = new ImapServerInfo(serverUri);
      this.id = id;
    }

    ~ImapConnection()
    {
      Disconnect();
    }

    static public ImapConnection Connect(ImapUri serverUri, ImapCreateAuthenticatedStreamCallback createAuthStreamCallback)
    {
      var connection = new ImapConnection(serverUri, Interlocked.Increment(ref connectionIdCount));

      if (connectionIdCount == int.MaxValue)
        Interlocked.Exchange(ref connectionIdCount, 0);

      connection.client = new TcpClient();

      if (serverUri.Port == -1)
        connection.client.Connect(serverUri.Host, ImapUri.DefaultPort);
      else
        connection.client.Connect(serverUri.Host, serverUri.Port);

      if (createAuthStreamCallback == null)
        connection.stream = connection.client.GetStream();
      else
        connection.stream = createAuthStreamCallback(serverUri, connection.client.GetStream());

      connection.receiver = new Receiver(connection.stream);
      //connection.sender = new Sender(connection.stream);

      // for first receiving
      connection.receiver.TryReceive();

      connection.connected = true;

      Trace.Info("CID:{0} connected to {1} ({2})", connection.Id, connection.client.Client.RemoteEndPoint, serverUri.Host);

      return connection;
    }

    public void Disconnect()
    {
      if (!connected)
        return;

      Trace.Info("CID:{0} disconnecting from {1}", id, client.Client.RemoteEndPoint);

      if (client != null) {
        client.Close();
        client = null;
      }

      Trace.Info("CID:{0} disconnected", id);

      connected = false;
    }

    public ImapProtocolCommand CreateCommand(string command, params string[] parameters)
    {
      var comm = new ImapProtocolCommand(command, commandTag.ToString("D4"), parameters);

      if (Interlocked.Increment(ref commandTag) == 0x10000)
        Interlocked.Exchange(ref commandTag, 0);

      return comm;
    }

    public ImapProtocolCommand CreateContinuingCommand(params string[] parameters)
    {
      return new ImapProtocolCommand(null, null, parameters);
    }

    public void SendCommand(ImapProtocolCommand command)
    {
      var comm = command.Format();

      Trace.Log(this, comm);

      stream.Write(comm, 0, comm.Length);

      lastSentTime = DateTime.Now;
    }

    public ImapProtocolResponse TryReceiveResponse()
    {
      // return parsed response
      if (0 < parsedResponse.Count)
        return parsedResponse.Dequeue();

      // get received data
      byte[] received = receiver.TryReceive();

      if (received == null)
        return null;

      // parse received data
      foreach (var response in parser.Parse(received)) {
        Trace.Log(this, response);
        Trace.Dump(response);

        parsedResponse.Enqueue(response);
      }

      // return parsed response
      if (0 < parsedResponse.Count)
        return parsedResponse.Dequeue();

      return null;
    }

    public override string ToString()
    {
      return string.Format("{{Host={0}, Connected={1}, Id={2}}}",
                           client.Client.RemoteEndPoint,
                           client.Client.Connected,
                           id);
    }

    private static int connectionIdCount = 0;
    private readonly int id = 0;
    private bool connected = false;
    private bool isIdling = false;
    private readonly ImapServerInfo serverInfo;
    private DateTime lastSentTime = DateTime.Now;
    private TcpClient client;
    private Stream stream;
    private int commandTag = 0;
    private ImapIdlingResponseCallback idlingCallback = null;
    private readonly ImapProtocolResponseParser parser = new ImapProtocolResponseParser();
    private readonly Queue<ImapProtocolResponse> parsedResponse = new Queue<ImapProtocolResponse>();
    private Receiver receiver = null;
    //private Sender sender = null;
  }
}
