using System;
using System.Collections.Generic;
using System.Text;
using System.Security.Cryptography;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction.Transactions {
  internal class AuthenticateTransaction : TransactionBase {

    // ProcessAuthenticateIncapable
    //   x Conclude
    // ProcessParameterNotSetted
    //   x Conclude
    // ProcessAuthenticate
    //   x Conclude
    //   ->WaitForAuthenticate
    // WaitForAuthenticate
    //   ContinuePlainAuthenticate
    //     ->ProcessAuthenticate
    //   ContinueLoginAuthenticate
    //     ->ProcessAuthenticate
    //   ContinueCramMd5Authenticate
    //     ->ProcessAuthenticate
    //   ContinueKerberosV4Authenticate
    //     ->ProcessAuthenticate
    //   x Conclude

    public AuthenticateTransaction(ImapConnection connection, ImapRequest request)
      : base(connection, request)
    {
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Request.Parameters.ContainsKey("Username"))
        return ProcessParameterNotSetted;

      username = Request.Parameters["Username"];

      if (Request.Parameters.ContainsKey("Password"))
        password = Request.Parameters["Password"];
      else
        password = string.Empty;

      if (Request.Parameters.ContainsKey("Auth-Type"))
        if (string.Empty.Equals(Request.Parameters["Auth-Type"]))
          tryingAuthType = "*";
        else
          tryingAuthType = Request.Parameters["Auth-Type"];
      else
        tryingAuthType = "*";

      trialAuthTypes.Clear();

      foreach (var auth in new[] {
        new {Cap = ImapKnownCapabilities.AuthKerberosV4, Type = "kerberos_v4" },
        new {Cap = ImapKnownCapabilities.AuthCramMd5,    Type = "cram-md5"},
        new {Cap = ImapKnownCapabilities.AuthPlain,      Type = "plain"},
        new {Cap = ImapKnownCapabilities.AuthLogin,      Type = "login"},
      }) {
        if ((Connection.ServerInfo.Capability.KnownCapabilities & auth.Cap) != 0)
          trialAuthTypes.Enqueue(auth.Type);
      }

      if (trialAuthTypes.Count == 0)
        return ProcessAuthenticateIncapable;

      return ProcessAuthenticate;
    }

    private void ProcessAuthenticateIncapable()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.Incapable, "AUTHENTICATE incapable"));
    }

    private void ProcessParameterNotSetted()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.BadRequest, "paramater 'Username' must be setted"));
    }

    private void ProcessAuthenticate()
    {
      if (tryingAuthType == "*") {
        // 認証方法に指定が無い場合は、CAPABILITYで返された認証方法から選ぶ
        if (trialAuthTypes.Count == 0) {
          // 他に使用できる認証方法がない場合は、失敗
          Conclude(new ImapResponse(ImapTransactionStatus.Forbidden, "login refused"));
          return;
        }
        else {
          // 優先順位が一番高いものを選ぶ
          tryingAuthType = trialAuthTypes.Dequeue();
        }
      }

      lastTag = Send("authenticate", tryingAuthType);

      Transit(WaitForAuthenticate);
    }

    private void WaitForAuthenticate()
    {
      ImapProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is ImapCommandContinuationRequest) {
        var continuationRequest = response as ImapCommandContinuationRequest;

        switch (tryingAuthType) {
          case "plain":
            ContinuePlainAuthenticate(continuationRequest);
            break;

          case "login":
            ContinueLoginAuthenticate(continuationRequest);
            break;

          case "cram-md5":
            ContinueCramMd5Authenticate(continuationRequest);
            break;

          case "kerberos_v4":
            ContinueKerberosV4Authenticate(continuationRequest);
            break;
        }
      }
      else if (response is ImapStatusResponse) {
        var status = (response as ImapStatusResponse);

        if (status.Tag != lastTag)
          return;

        if (status.Status == ImapProtocolStatus.OK) {
          Conclude(new ImapResponse(ImapTransactionStatus.Ok, "logged in")); 
        }
        else {
          // 認証出来なかった場合、他の可能な手段で認証する
          tryingAuthType = "*";

          Transit(ProcessAuthenticate);
        }
      }
    }

    private void ContinuePlainAuthenticate(ImapCommandContinuationRequest continuationRequest)
    {
      var plainAuthString = string.Format("\0{0}\0{1}", username, password);

      plainAuthString = Convert.ToBase64String(Encoding.ASCII.GetBytes(plainAuthString));

      SendContinue(plainAuthString);
    }

    private void ContinueLoginAuthenticate(ImapCommandContinuationRequest continuationRequest)
    {
      var prompt = Encoding.ASCII.GetString(Convert.FromBase64String(continuationRequest.Text));
      
      switch (prompt) {
        case "Username:":
          SendContinue(Convert.ToBase64String(Encoding.ASCII.GetBytes(username)));
          break;
        
        case "Password:":
          SendContinue(Convert.ToBase64String(Encoding.ASCII.GetBytes(password)));
          break;

        default:
          // ?
          SendContinue(string.Empty);
          break;
      }
    }

    private void ContinueCramMd5Authenticate(ImapCommandContinuationRequest continuationRequest)
    {
      // RFC 2195 IMAP/POP AUTHorize Extension for Simple Challenge/Response
      // http://tools.ietf.org/html/rfc2195

      // 2. Challenge-Response Authentication Mechanism (CRAM)

      // The data encoded in the first ready response contains an
      // presumptively arbitrary string of random digits, a timestamp, and the
      // fully-qualified primary host name of the server.  The syntax of the
      // unencoded form must correspond to that of an RFC 822 'msg-id'
      // [RFC822] as described in [POP3].
      var challenge = Convert.FromBase64String(continuationRequest.Text);

      // The client makes note of the data and then responds with a string
      // consisting of the user name, a space, and a 'digest'.  The latter is
      // computed by applying the keyed MD5 algorithm from [KEYED-MD5] where
      // the key is a shared secret and the digested text is the timestamp
      // (including angle-brackets).
      //
      // This shared secret is a string known only to the client and server.
      var keyed = (new HMACMD5(Encoding.ASCII.GetBytes(password))).ComputeHash(challenge);

      // The `digest' parameter itself is a 16-octet value which is sent in
      // hexadecimal format, using lower-case ASCII characters.
      var digest = string.Empty;

      foreach (var octet in keyed) {
        digest += octet.ToString("x02");
      }

      var resp = string.Format("{0} {1}", username, digest);

      SendContinue(Convert.ToBase64String(Encoding.ASCII.GetBytes(resp)));
    }

    private void ContinueKerberosV4Authenticate(ImapCommandContinuationRequest continuationRequest)
    {
      // TODO: implement
      SendContinue(string.Empty);
    }

    private string lastTag = string.Empty;
    private string username = string.Empty;
    private string password = string.Empty;
    private Queue<string> trialAuthTypes = new Queue<string>();
    private string tryingAuthType;
  }
}
