using System;
using System.Collections.Generic;
using System.Threading;

using Smdn.Protocols.Pop3.Protocol;

namespace Smdn.Protocols.Pop3.Transaction {
  // Layers:
  //   Client.PopClient
  //   Client.PopClientAgent
  //   Session.PopSessionAgent
  //   Session.PopSession
  // * Transaction.PopTransactionAgent
  //   Transaction.PopTransaction/Transaction.PopRequest/Transaction.PopResponse
  //   Protocol.PopConnection/Protocol.PopProtocolCommand/Protocol.PopProtocolResponse
  //   (POP3)

  public class PopTransactionAgent : IDisposable {
    /*
     * construction/destruction
     */

    public PopTransactionAgent()
    {
    }

    public void Initialize()
    {
      if (agentThread != null)
        throw new InvalidOperationException("already initialized");

      Trace.Info("PopTransactionAgent initializing");

      agentThread = new Thread((ThreadStart)AgentThread);
      agentThread.IsBackground = true;
      agentThread.Start();

      // wait for initial WaitAny blocking
      for (;; Thread.Sleep(100)) {
        if ((agentThread.ThreadState & ThreadState.WaitSleepJoin) != 0)
          break;
      }

      Trace.Info("PopTransactionAgent initialized");
    }

    ~PopTransactionAgent()
    {
      Dispose();
      GC.SuppressFinalize(this);
    }

    public void Dispose()
    {
      if (disposed)
        return;

      Trace.Info("PopTransactionAgent disposing");

      abortedEvent.Set();

      if (agentThread != null) {
        agentThread.Join();
        agentThread = null;
      }

      Trace.Info("PopTransactionAgent disposed");

      disposed = true;
    }

    /*
     * methods for requesting
     */

    public PopTransaction Request(PopConnection connection, PopRequest request)
    {
      if (abortedEvent.WaitOne(0, false))
        throw new InvalidOperationException("aborting");

      var transaction = PopTransaction.Create(connection, request);

      transaction.Request.RequestedTime = DateTime.Now;

      // check whether another transaction on this connection is in progress
      foreach (var t in processingTransactions.ToArray()) {
        if (t.Connection == connection) {
          transaction.Response = new PopResponse(PopTransactionStatus.TransactionProgress,
                                                "another transaction on this connection is in progress");
          return transaction;
        }
      }

      // enqueue
      lock (processingTransactions) {
        processingTransactions.Add(transaction);
      }

      Trace.Log(request, transaction);

      Signal();

      return transaction;
    }

    /// <summary>sends a activation signal to transaction processing thread</summary>
    public void Signal()
    {
      requestedEvent.Set();
    }

    /*
     * methods for transaction processing thread
     */

    private void AgentThread()
    {
      Trace.Info("PopTransactionAgent thread started");

      try {
        var waitEvents = new WaitHandle[] {abortedEvent, requestedEvent};

        abortedEvent.Reset();

        // wait for first request (or abort request)
        WaitHandle.WaitAny(waitEvents); 

        for (;;) {
          PopTransaction[] transactions;

          lock (processingTransactions) {
            transactions = processingTransactions.ToArray();
          }

          bool aborted;

          ProcessTransactions(transactions, out aborted);

          if (aborted)
            break;

          if (0 < processingTransactions.Count) {
            // トランザクションがある場合は、すぐに処理する
            Thread.Sleep(10);
          }
          else {
            // 処理すべきトランザクションが無い場合は、リクエストが来るか、
            // スレッドの停止が要求されるまで待機する
            var index = WaitHandle.WaitAny(waitEvents); 

            if (index == 0)
              // aborted
              break;
          }
        }

        // abort all transactions
        foreach (var transaction in processingTransactions) {
          transaction.Abort();
        }

        processingTransactions.Clear();
      }
      catch (ThreadAbortException) {
        Trace.Info("PopTransactionAgent thread force aborted");
      }
      catch (Exception ex) {
        Trace.Log(ex);
        throw ex;
      }

      Trace.Info("PopTransactionAgent thread finished");
    }

    private void ProcessTransactions(PopTransaction[] transactions, out bool aborted)
    {
      aborted = false;

      foreach (var transaction in transactions) {
        // aborted
        if (abortedEvent.WaitOne(0, false)) {
          aborted = true;
          return;
        }

        if (ProcessTransaction(transaction)) {
          // transaction is finished
          lock (processingTransactions) {
            processingTransactions.Remove(transaction);
          }
        }
      }
    }

    private bool ProcessTransaction(PopTransaction transaction)
    {
      transaction.Process();

      if (transaction.Response != null) {
        // transaction is finished
        Trace.Log(transaction.Response, transaction);
        return true;
      }

      // transaction is not finished
      if (transaction.Request.Timeout < 0)
        return false;

      if (transaction.Request.Timeout <
          DateTime.Now.Subtract(transaction.Request.RequestedTime).TotalMilliseconds)
        // request time out
        transaction.Abort();

      return false;
    }

    private bool disposed = false;
    private readonly List<PopTransaction> processingTransactions = new List<PopTransaction>();
    private Thread agentThread = null;
    private readonly AutoResetEvent requestedEvent = new AutoResetEvent(false);
    private readonly ManualResetEvent abortedEvent = new ManualResetEvent(false);
  }
}
