using System;
using System.Collections.Generic;
using System.Text;
using System.Security.Cryptography;

using Smdn.Protocols.Pop3.Protocol;

namespace Smdn.Protocols.Pop3.Transaction.Transactions {
  internal class AuthTransaction : TransactionBase {

    // ProcessAuthIncapable
    //   x Conclude
    // ProcessParameterNotSetted
    //   x Conclude
    // ProcessAuth
    //   x Conclude
    //   ->WaitForAuth
    // WaitForAuth
    //   ContinuePlainAuthenticate
    //     ->ProcessAuthenticate
    //   ContinueLoginAuthenticate
    //     ->ProcessAuthenticate
    //   ContinueCramMd5Authenticate
    //     ->ProcessAuthenticate
    //   ContinueKerberosV4Authenticate
    //     ->ProcessAuthenticate
    //   x Conclude

    public AuthTransaction(PopConnection connection, PopRequest request)
      : base(connection, request)
    {
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Request.Parameters.ContainsKey("Mailbox"))
        return ProcessParameterNotSetted;

      mailbox = Request.Parameters["Mailbox"];

      if (Request.Parameters.ContainsKey("Password"))
        password = Request.Parameters["Password"];
      else
        password = string.Empty;

      if (Request.Parameters.ContainsKey("Auth-Type"))
        if (string.Empty.Equals(Request.Parameters["Auth-Type"]))
          tryingAuthType = "*";
        else
          tryingAuthType = Request.Parameters["Auth-Type"];
      else
        tryingAuthType = "*";

      trialAuthTypes.Clear();

      foreach (var auth in new[] {
        new {Cap = PopKnownCapabilities.AuthKerberosV4, Type = "KERBEROS_V4" },
        new {Cap = PopKnownCapabilities.AuthCramMd5,    Type = "CRAM-MD5"},
        new {Cap = PopKnownCapabilities.AuthPlain,      Type = "PLAIN"},
        new {Cap = PopKnownCapabilities.AuthLogin,      Type = "LOGIN"},
      }) {
        if ((Connection.ServerInfo.Capability.KnownCapabilities & auth.Cap) != 0)
          trialAuthTypes.Enqueue(auth.Type);
      }

      if ((Connection.ServerInfo.Capability.KnownCapabilities & PopKnownCapabilities.Capabilities) != 0 &&
          trialAuthTypes.Count == 0)
        return ProcessAuthIncapable;
        
      return ProcessAuth;
      
    }

    private void ProcessAuthIncapable()
    {
      Conclude(new PopResponse(PopTransactionStatus.Incapable, "AUTH incapable"));
    }

    private void ProcessParameterNotSetted()
    {
      Conclude(new PopResponse(PopTransactionStatus.BadRequest, "parameter 'Mailbox' must be setted"));
    }

    private void ProcessAuth()
    {
      if (tryingAuthType == "*") {
        // 認証方法に指定が無い場合は、CAPAで返された認証方法から選ぶ
        if (trialAuthTypes.Count == 0) {
          // 他に使用できる認証方法がない場合は、失敗
          Conclude(new PopResponse(PopTransactionStatus.Forbidden, "AUTH refused"));
          return;
        }
        else {
          // 優先順位が一番高いものを選ぶ
          tryingAuthType = trialAuthTypes.Dequeue();
        }
      }

      Send("AUTH", tryingAuthType);

      Transit(WaitForAuth);
    }

    private void WaitForAuth()
    {
      PopProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is PopContinuationRequest) {
        var continuationRequest = response as PopContinuationRequest;

        switch (tryingAuthType) {
          case "PLAIN":
            ContinuePlainAuth(continuationRequest);
            break;

          case "LOGIN":
            ContinueLoginAuth(continuationRequest);
            break;

          case "CRAM-MD5":
            ContinueCramMd5Auth(continuationRequest);
            break;

          case "KERBEROS_V4":
            ContinueKerberosV4Auth(continuationRequest);
            break;
        }
      }
      else if (response is PopStatusResponse) {
        var status = (response as PopStatusResponse);

        if (status.Status == PopProtocolStatus.Positive) {
          Conclude(new PopResponse(PopTransactionStatus.Ok, "AUTH succeeded")); 
        }
        else {
          // 認証出来なかった場合、他の可能な手段で認証する
          tryingAuthType = "*";

          Transit(ProcessAuth);
        }
      }
    }

    private void ContinuePlainAuth(PopContinuationRequest continuationRequest)
    {
      var plainAuthString = string.Format("\0{0}\0{1}", mailbox, password);

      Send(Convert.ToBase64String(Encoding.ASCII.GetBytes(plainAuthString)));
    }

    private void ContinueLoginAuth(PopContinuationRequest continuationRequest)
    {
      switch (continuationRequest.Text) {
        case "Username:":
          Send(Convert.ToBase64String(Encoding.ASCII.GetBytes(mailbox)));
          break;
        
        case "Password:":
          Send(Convert.ToBase64String(Encoding.ASCII.GetBytes(password)));
          break;

        default:
          // ?
          Send(string.Empty);
          break;
      }
    }

    private void ContinueCramMd5Auth(PopContinuationRequest continuationRequest)
    {
      // RFC 2195 IMAP/POP AUTHorize Extension for Simple Challenge/Response
      // http://tools.ietf.org/html/rfc2195

      // 2. Challenge-Response Authentication Mechanism (CRAM)

      // The data encoded in the first ready response contains an
      // presumptively arbitrary string of random digits, a timestamp, and the
      // fully-qualified primary host name of the server.  The syntax of the
      // unencoded form must correspond to that of an RFC 822 'msg-id'
      // [RFC822] as described in [POP3].
      var challenge = Encoding.ASCII.GetBytes(continuationRequest.Text);

      // The client makes note of the data and then responds with a string
      // consisting of the user name, a space, and a 'digest'.  The latter is
      // computed by applying the keyed MD5 algorithm from [KEYED-MD5] where
      // the key is a shared secret and the digested text is the timestamp
      // (including angle-brackets).
      //
      // This shared secret is a string known only to the client and server.
      var keyed = (new HMACMD5(Encoding.ASCII.GetBytes(password))).ComputeHash(challenge);

      // The `digest' parameter itself is a 16-octet value which is sent in
      // hexadecimal format, using lower-case ASCII characters.
      var digest = string.Empty;

      foreach (var octet in keyed) {
        digest += octet.ToString("x02");
      }

      var resp = string.Format("{0} {1}", mailbox, digest);

      Send(Convert.ToBase64String(Encoding.ASCII.GetBytes(resp)));
    }

    private void ContinueKerberosV4Auth(PopContinuationRequest continuationRequest)
    {
      // TODO: implement
      Send(string.Empty);
    }

    private string mailbox = string.Empty;
    private string password = string.Empty;
    private Queue<string> trialAuthTypes = new Queue<string>();
    private string tryingAuthType;
  }
}
