using System;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Reflection;
using System.Threading;

using Smdn.Protocols.Pop3.Protocol;
using Smdn.Protocols.Pop3.Transaction;

namespace Smdn.Protocols.Pop3 {
  internal static class Trace {
    private static readonly TraceSource     traceSource = new TraceSource(Assembly.GetExecutingAssembly().GetName().Name, SourceLevels.Verbose);
    private static readonly TraceSource dumpTraceSource = new TraceSource(Assembly.GetExecutingAssembly().GetName().Name + "#Dump", SourceLevels.Verbose);

    private static string Now()
    {
      return string.Format("{0:s}.{1:d4} ", DateTime.Now, DateTime.Now.Millisecond);
    }

    [Conditional("TRACE")]
    public static void Info(string format, params object[] args)
    {
      traceSource.TraceEvent(TraceEventType.Information, Thread.CurrentThread.ManagedThreadId, Now() + format, args);
    }

    [Conditional("TRACE")]
    public static void Verbose(string format, params object[] args)
    {
      traceSource.TraceEvent(TraceEventType.Verbose, Thread.CurrentThread.ManagedThreadId, Now() + format, args);
    }

    [Conditional("TRACE")]
    public static void Error(string format, params object[] args)
    {
      traceSource.TraceEvent(TraceEventType.Error, Thread.CurrentThread.ManagedThreadId, Now() + format, args);
    }

    [Conditional("TRACE")]
    public static void Log(Exception ex)
    {
      traceSource.TraceEvent(TraceEventType.Error, Thread.CurrentThread.ManagedThreadId, "Σ（＞ω＜）{0}{1}", Environment.NewLine, ex);
    }

    [Conditional("TRACE")]
    public static void Log(PopConnection connection, PopProtocolResponse response)
    {
      var prefix = string.Format("CID:{0} 　Σ（∵）[POP3]", connection.Id);

      var reader = new StringReader(response.Response);

      for (;;) {
        var line = reader.ReadLine();

        if (line == null)
          break;

        Verbose("{0} {1}", prefix, line);
      }
    }

    [Conditional("TRACE")]
    public static void Log(PopConnection connection, byte[] command)
    {
      var prefix = string.Format("CID:{0} （＞ω＜）[POP3]", connection.Id);

      var reader = new StringReader(Encoding.ASCII.GetString(command));

      for (;;) {
        var line = reader.ReadLine();

        if (line == null)
          break;

        Verbose("{0} {1}", prefix, line);
      }
    }

    [Conditional("TRACE")]
    public static void Log(PopRequest request, PopTransaction transaction)
    {
      var param = string.Empty;

      foreach (var p in request.Parameters) {
        var val = p.Value;

        if (p.Key == "Password")
          val = "****";

        param += string.Format(" {0}={1}", p.Key, val);
      }

      Verbose("CID:{0} SeqID:{1} [PopTransaction] {2}{3}",
              transaction.Connection.Id,
              transaction.SequenceId,
              request.Method,
              param);
    }

    [Conditional("TRACE")]
    public static void Log(PopResponse response, PopTransaction transaction)
    {
      Verbose("CID:{0} SeqID:{1} [PopTransaction] {2} {3} {4}",
              transaction.Connection.Id,
              transaction.SequenceId,
              transaction.Request.Method,
              response.Status,
              response.Description);
    }

    [Conditional("DUMP")]
    public static void DumpInfo(string format, params object[] args)
    {
      dumpTraceSource.TraceEvent(TraceEventType.Information, Thread.CurrentThread.ManagedThreadId, Now() + format, args);
    }

    [Conditional("DUMP")]
    public static void DumpVerbose(string format, params object[] args)
    {
      dumpTraceSource.TraceEvent(TraceEventType.Verbose, Thread.CurrentThread.ManagedThreadId, Now() + format, args);
    }

    [Conditional("DUMP")]
    public static void Dump(byte[] buffer)
    {
      var sb = new StringBuilder();

      for (var index = 0; index < buffer.Length; index++) {
        if (index % 0x20 == 0)
          sb.AppendLine();
        sb.AppendFormat("{0:x2} ", buffer[index]);
      }

      DumpVerbose("Length:{0}{1}{2}{3}",
              buffer.Length,
              Environment.NewLine,
              Encoding.ASCII.GetString(buffer),
              sb.ToString());
    }
  }
}
